@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2014-2017 Savoir-faire Linux Inc.
+ *  Copyright (C) 2014-2022 Savoir-faire Linux Inc.
  *  Author : Adrien Béraud <adrien.beraud@savoirfairelinux.com>
  *
  *  This program is free software; you can redistribute it and/or modify
@@ -21,6 +21,8 @@
 #include <random>
 #include <algorithm>
 #include <functional>
+#include <thread>
+#include <stdexcept>
 
 namespace dht {
 namespace crypto {
@@ -43,8 +45,8 @@ public:
      * result_type is unsigned int so this is currently safe.
      */
     static_assert(
-        sizeof(result_type) == 2 || 
-        sizeof(result_type) == 4 || 
+        sizeof(result_type) == 2 ||
+        sizeof(result_type) == 4 ||
         sizeof(result_type) == 8,
         "result_type must be 16, 32 or 64 bits");
 
@@ -111,11 +113,18 @@ using random_device = std::random_device;
 template<class T = std::mt19937, std::size_t N = T::state_size>
 auto getSeededRandomEngine () -> typename std::enable_if<!!N, T>::type {
     typename T::result_type random_data[N];
-    random_device source;
-    std::generate(std::begin(random_data), std::end(random_data), std::ref(source));
-    std::seed_seq seeds(std::begin(random_data), std::end(random_data));
-    T seededEngine (seeds);
-    return seededEngine;
+    for (unsigned i=0; i<256; i++) {
+        try {
+            random_device source;
+            std::generate(std::begin(random_data), std::end(random_data), std::ref(source));
+            std::seed_seq seeds(std::begin(random_data), std::end(random_data));
+            T seededEngine (seeds);
+            return seededEngine;
+        } catch (...) {
+            std::this_thread::sleep_for(std::chrono::milliseconds(1));
+        }
+    }
+    throw std::runtime_error("Can't seed random engine");
 }
 
 }} // dht::crypto
