@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2014-2020 Savoir-faire Linux Inc.
+ *  Copyright (C) 2014-2022 Savoir-faire Linux Inc.
  *
  *  Author: Adrien Béraud <adrien.beraud@savoirfairelinux.com>
  *  Author: Sébastien Blin <sebastien.blin@savoirfairelinux.com>
@@ -131,8 +131,10 @@ struct dht_params {
     in_port_t proxyserver {0};
     in_port_t proxyserverssl {0};
     std::string proxyclient {};
+    std::string proxy_address {};
     std::string pushserver {};
     std::string devicekey {};
+    std::string bundle_id {};
     std::string persist_path {};
     dht::crypto::Identity id {};
     dht::crypto::Identity proxy_id {};
@@ -223,12 +225,14 @@ static const constexpr struct option long_options[] = {
     {"syslog",                  no_argument      , nullptr, 'L'},
     {"proxyserver",             required_argument, nullptr, 'S'},
     {"proxyserverssl",          required_argument, nullptr, 'e'},
+    {"proxy-addr",              required_argument, nullptr, 'a'},
     {"proxy-certificate",       required_argument, nullptr, 'w'},
     {"proxy-privkey",           required_argument, nullptr, 'K'},
     {"proxy-privkey-password",  required_argument, nullptr, 'M'},
     {"proxyclient",             required_argument, nullptr, 'C'},
     {"pushserver",              required_argument, nullptr, 'y'},
     {"devicekey",               required_argument, nullptr, 'z'},
+    {"bundleid",                required_argument, nullptr, 'u'},
     {"version",                 no_argument      , nullptr, 'V'},
     {nullptr,                   0                , nullptr,  0}
 };
@@ -239,7 +243,7 @@ parseArgs(int argc, char **argv) {
     int opt;
     std::string privkey;
     std::string proxy_privkey;
-    while ((opt = getopt_long(argc, argv, "hidsvDUPp:n:b:f:l:", long_options, nullptr)) != -1) {
+    while ((opt = getopt_long(argc, argv, "hidsvODUPp:n:b:f:l:", long_options, nullptr)) != -1) {
         switch (opt) {
         case 'p': {
                 int port_arg = atoi(optarg);
@@ -268,6 +272,9 @@ parseArgs(int argc, char **argv) {
         case 'D':
             params.peer_discovery = true;
             break;
+        case 'a':
+            params.proxy_address = optarg;
+            break;
         case 'y':
             params.pushserver = optarg;
             break;
@@ -336,6 +343,9 @@ parseArgs(int argc, char **argv) {
         case 'k':
             privkey = optarg;
             break;
+        case 'u':
+            params.bundle_id = optarg;
+            break;
         case 'K':
             proxy_privkey = optarg;
             break;
